//
// Created by 320051849 on 26/02/2022.
//

#include <Arduino.h>
#include "terminal.h"

Terminal::Terminal(LCDWIKI_KBV *lcd, uint8_t lines, uint8_t chars, uint16_t xLoc, uint16_t yLoc, uint16_t bg, uint16_t fg)
{
    // Constructor stuff here
    _lcd    = lcd;
    _lines  = lines;
    _chars  = chars;
    _xLoc   = xLoc;
    _yLoc   = yLoc;
    _bg     = bg;
    _fg     = fg;

    State   = Line_Update;
}

void Terminal::WriteString(const char *str, uint16_t x, uint16_t y, uint8_t csize, uint16_t fc, uint16_t bc, boolean mode) 
{
    _lcd->Set_Text_Mode(mode);
    _lcd->Set_Text_Size(csize);
    _lcd->Set_Text_colour(fc);
    _lcd->Set_Text_Back_colour(bc);
    _lcd->Print_String(str,(int16_t) x,(int16_t) y);
}

void Terminal::WriteTerminalLine(const char *str) 
{
    static int  newest_line = 1;
    static int  oldest_line = 0;
    static bool buffer_full = false;

    // Copy input string to 'newest' slot in terminal _buffer, update _indexing variable
    snprintf(_buffer[newest_line], _chars,"%s", str);
    _index = newest_line;

    // Update indices
    newest_line++;
    //_index = newest_line;
    if (newest_line == _lines)
    {
        newest_line = 0;
    }
    if (buffer_full)
    {
        oldest_line++;
        if (oldest_line == _lines)
        {
            oldest_line = 0;
        }
    }
    // Set state indicating that a line has updated, thus the entire console must be redrawn
    State = Line_Update;
}

void Terminal::WriteTerminalChars(const char *str) 
{
    char     chars_buffer[BUFFER_CHARS] = "\0";
    uint16_t num_new_chars = strlen(str);
    uint16_t num_old_chars = strlen(_buffer[_index]);
    uint16_t freespace, charswritten = 0;
    uint16_t charsleft = num_new_chars;
    uint16_t i;

    // If less than the maximum allowed, simply append characters to existing _buffer contents
    if ((num_old_chars + num_new_chars) < (_chars))
    {
        strcat(_buffer[_index], str);
        State = Chars_Update;
        if (str[num_new_chars-1] == '\r')
        {
            strcpy(chars_buffer, _buffer[_index]);
            WriteTerminalLine("");
            Serial.println("");
        }
    }
        // If more than the maximum allowed, append the maximum characters and load the _buffer with the overflow
    else
    {
        //overflow = (num_new_chars + num_old_chars + 1 - _chars);
        freespace = (_chars - num_old_chars - 1);
        strcpy(chars_buffer, _buffer[_index]);
        for (i = 0; (i < freespace); i++)
        {
            chars_buffer[num_old_chars+i] = str[i];
            charsleft--;
            charswritten++;
        }
        chars_buffer[_chars-1] = '\0';
        strcpy(_buffer[_index], chars_buffer);  // Todo: remove intermediary _buffer (
        WriteTerminalLine("");
        for (i = 0; i < charsleft+1; i++)
        {
            chars_buffer[i] = str[charswritten+i];
        }
        chars_buffer[charsleft] = '\0';
        strcpy(_buffer[_index], chars_buffer);
    }
}

void Terminal::DisplayTerminal()
{
    uint16_t x = _xLoc;
    uint16_t y = _yLoc;
    int i;
    char line_buffer[_chars+2];

    // Full update of entire terminal
    if (State == Line_Update)
    {
        // Clear text output area before (re)drawing terminal output & reposition Y start location
        _lcd->Fill_Rect((int16_t)x, (int16_t)y, (6 * (_chars + 1)), (10 * _lines), _bg);
        y += 10 * (_lines - 1);

        // Write prompt
        WriteString(">", x, y, 1, _fg, _bg, true);
        for (i=_index; i>=0; i--)
        {
            snprintf(line_buffer, _chars+2,"%s", _buffer[i]);
            WriteString(line_buffer, (x + 10), y, 1, _fg, _bg, true);
            y -= 10;
        }
        for (i=_lines-1; i>_index; i--)
        {
            snprintf(line_buffer, _chars+2,"%s", _buffer[i]);
            WriteString(line_buffer, (x + 10), y, 1, _fg, _bg, true);
            y -= 10;
        }
    }
        // Just the current line of the _buffer to be updated
    else if (State == Chars_Update)
    {
        y += 10 * (_lines-1);
        snprintf(line_buffer, _chars+2,"%s", _buffer[_index]);
        WriteString(line_buffer, (x + 10), y, 1, _fg, _bg, true);
    }
    // Reset state
    State = No_Update;
}

void Terminal::DisplayBuffer(uint16_t xLoc, uint16_t yLoc) 
{
    char line_buffer[_chars+2];
    int  i;

    // Clear text output area before redrawing _buffer contents
    _lcd->Fill_Rect((int16_t)xLoc, (int16_t)yLoc, (6*(_chars+2)), (10*_lines), _bg);
    //_lcd.Fill_Round_Rectangle(6, 6, (_lcd_size_x/2-1), (_lcd_size_y-7), 7);

    for (i=0; i<_lines; i++)
    {
        snprintf(line_buffer, _chars+2,"%i %s", i, _buffer[i]);
        WriteString(line_buffer, xLoc, yLoc, 1, _fg, _bg, true);
        yLoc += 10;
    }
}
