/***********************************************************************************
* Smart Charger Controller (For Arduino Mega)
* Written by GLO
* Last modified: 03/03/2022
**********************************************************************************/

#include <Arduino.h>
//#include <UTFT.h>
//#include <UTFTGLUE.h>
#include <LCDWIKI_KBV.h>
#include <LCDWIKI_GUI.h>
//#include <TouchScreen.h>    // Touchscreen support
#include <Wire.h>           // I2C support
#include <CRC.h>            // CRC calculations
#include <terminal.h>       // Terminal (GMLO)

#define TITLE       "Smart Charger Controller (For Arduino Mega)"

/*
// Touchscreen Defines
#define YP A3               // must be an analog pin, use "An" notation!
#define XM A2               // must be an analog pin, use "An" notation!
#define YM 9                // can be a digital pin
#define XP 8                // can be a digital pin
#define TS_MINX 945
#define TS_MAXX 175
#define TS_MINY 230
#define TS_MAXY 813
#define MINPRESSURE 10
#define MAXPRESSURE 1000
*/

// Standard and common non-standard Smart Battery commands
#define BATTERY_MODE             0x03
#define TEMPERATURE              0x08
#define VOLTAGE                  0x09
#define CURRENT                  0x0A
#define RELATIVE_SOC             0x0D
#define ABSOLUTE_SOC             0x0E
#define REMAINING_CAPACITY       0x0F
#define FULL_CHARGE_CAPACITY     0x10
#define TIME_TO_FULL             0x13
#define CHARGING_CURRENT         0x14
#define CHARGING_VOLTAGE         0x15
#define BATTERY_STATUS           0x16
#define CYCLE_COUNT              0x17
#define DESIGN_CAPACITY          0x18
#define DESIGN_VOLTAGE           0x19
#define SPEC_INFO                0x1A
#define MFG_DATE                 0x1B
#define SERIAL_NUM               0x1C
#define MFG_NAME                 0x20   // String
#define DEV_NAME                 0x21   // String
#define CELL_CHEM                0x22   // String
#define MFG_DATA                 0x23   // String
#define MFG_DATA3B               0x4F   // String
//#define MFG_DATA3C               0x3C   // String
//#define MFG_DATA3D               0x3D   // String
//#define MFG_DATA3E               0x3E   // String
//#define MFG_DATA30               0x48   // String
//#define MFG_DATA31               0x49   // String
//#define MFG_DATA32               0x4A   // String
//#define MFG_DATA33               0x4B   // String
#define CELL8_VOLTAGE            0x43
#define CELL7_VOLTAGE            0x42
#define CELL6_VOLTAGE            0x41
#define CELL5_VOLTAGE            0x40
#define CELL4_VOLTAGE            0x3C   // Indidual cell voltages don't work on Lenovo and Dell Packs
#define CELL3_VOLTAGE            0x3D
#define CELL2_VOLTAGE            0x3E
#define CELL1_VOLTAGE            0x3F
#define STATE_OF_HEALTH          0x54
#define bufferLen 32
// Address of BQ40Z50 IC on SMBus (7 bits)
byte address_bms = 0x0B;


#define BG_COLOUR 0x001F
#define FG_COLOUR 0xFFFF

// Defined states for main state machine
typedef enum {
    Update_Terminal = 0,
    Wait_For_Serial = 1,
    Wait_For_Touch = 2,
    Read_From_I2C = 3
    //State4 = 4,
    //State5 = 5
} Main_state_type;

// Class object for LCD
//LCDWIKI_KBV lcd(ILI9341, A3, A2, A1, A0, A4); // Controller, CS, RS, WR, RD, RST
LCDWIKI_KBV lcd(ILI9486, A3, A2, A1, A0, A4); // Controller, CS, RS, WR, RD, RST
//UTFTGLUE lcd(0,A2,A1,A3,A4,A0); //all dummy args
//UTFT lcd(ILI9486,A2,A1,A3,A4);
// Class object for terminal
Terminal terminal(&lcd, BUFFER_LINES, BUFFER_CHARS, 10, 50, 0x0000, 0xFFFF);
// Touchscreen and pointer objects
//TouchScreen ts(XP, YP, XM, YM, 300);
//TSPoint tp;                      //Touchscreen_due branch uses Point

// Global variables
Main_state_type CurrentState = Wait_For_Serial;
int  lcd_size_x, lcd_size_y;
uint32_t cx, cy, x, y;
//uint32_t rx[8], ry[8];
//uint8_t i2c_buffer[bufferLen];

/*
void Read_Resistive()
{
    tp = ts.getPoint();
    pinMode(YP, OUTPUT);      //restore shared pins
    pinMode(XM, OUTPUT);
    digitalWrite(YP, HIGH);     //because TFT control pins
    digitalWrite(XM, HIGH);

    if (tp.z > MINPRESSURE && tp.z < MAXPRESSURE) {
        //x = lcd_size_x-map(tp.x, TS_MINX, TS_MAXX, lcd_size_x, 0);
        //y = lcd_size_y-map(tp.y, TS_MINY, TS_MAXY, lcd_size_y, 0);
        x = map(tp.x, TS_MINX, TS_MAXX, 0, lcd_size_x);
        y = map(tp.y, TS_MINY, TS_MAXY, 0, lcd_size_y);
    }
}

bool is_pressed()
{
    bool state;
    //Read_Resistive();
    state = (tp.z > 20 && tp.z < 1000);
    return state;
}

void showpoint()
{
    Serial.print("\r\nx="); Serial.print(tp.x);
    Serial.print(" y="); Serial.print(tp.y);
    Serial.print(" z="); Serial.print(tp.z);
}

int16_t fetchWord(byte address)
{
    char temp_str[64] = "";
    byte byte_buffer[2] = {0, 0};
    byte byte_counter = 0;
    int  read_word;
    int  bytes_available;

    sprintf(temp_str, "  Requesting word from address 0x%02X...", address);
    Serial.println(temp_str);
    //Serial.print("  Response: 0x");

    bytes_available = Wire.available();
    Serial.print("  Bytes available to read: ");
    Serial.println(bytes_available);

    // Send request for 2x bytes
    Wire.requestFrom(address, (byte) 2, (byte) true);      // !Don't release the bus until complete
    while (bytes_available)
    {
        byte_buffer[byte_counter] = (byte) Wire.read();     // receive a byte
        Serial.print("    Byte read: 0x");
        Serial.println(byte_buffer[byte_counter], HEX);     // print the byte
        Serial.print("    Bytes left: ");
        bytes_available = Wire.available();
        Serial.println(bytes_available);
        byte_counter++;
    }

    read_word = (int)byte_buffer[1] | (((int) byte_buffer[0]) << 8);

    sprintf(temp_str, "  Word read from I2C: 0x%02X", read_word);
    Serial.println(temp_str);

    return read_word;
}

void sendByte(byte address, byte command)
{
    char temp_str[64] = "";

    sprintf(temp_str, "  Sending command 0x%02X to address 0x%02X", command, address);
    Serial.println(temp_str);

    // Write command
    Wire.beginTransmission(address);
    Wire.write(command);
    Wire.endTransmission();
    //delay(5);
}

void sendBytePEC(byte address, byte command)
{
    char temp_str[64] = "";
    byte frame[6] = {(byte) (address << 1), 0x44, 0x02, command, 0, 0};
    byte i;

    sprintf(temp_str, "  Sending command 0x%02X00 to address 0x%02X, PEC enabled", command, address);
    Serial.println(temp_str);

    // Apply CRC8 to last byte
    frame[5] = crc8(frame, 5, 0x07);

    // Write frame to I2C
    Serial.println(frame[0], HEX);
    Wire.beginTransmission(address);
    for (i = 1; i < 6; i++)
    {
        Wire.write(frame[i]);
        Serial.println(frame[i], HEX);
    }
    Wire.endTransmission();
    //delay(5);
}
*/
void setup()
{
    Serial.begin(9600);
    Serial.println("Serial comms initialized.");
    Serial.println(TITLE);
    // Init I2C as Master
    //Wire.begin();
    //Serial.println("I2C Inialized");
    // Init LCD
    //digitalWrite(A0, HIGH);             // LCD_RD Pin
    //pinMode(A0, OUTPUT);
    lcd.Init_LCD();
    //lcd.Fill_Screen(0x00);
    //lcd.InitLCD();
    //lcd.clrScr();
    //lcd.fillScreen(0x001F);               // Blue
    //lcd.reset();
    Serial.println(lcd.Read_ID(), HEX);
    Serial.println("LCD Inialized");
    lcd.Set_Rotation(3);                     // Portrait mode = 0, Landscape = 1
    lcd.Fill_Screen(BG_COLOUR);

    lcd_size_x = lcd.Get_Display_Width();
    lcd_size_y = lcd.Get_Display_Height();
    lcd.Set_Text_colour(FG_COLOUR);
    lcd.Set_Text_Back_colour(BG_COLOUR);
    // Debug LCD info to serial port
    Serial.print("Controller IC: 0x");
    Serial.println(lcd.Read_ID(), HEX);
    Serial.print("Display Width: ");
    Serial.println(lcd_size_x);
    Serial.print("Display Height: ");
    Serial.println(lcd_size_y);
    // Draw frame on left & label
    lcd.Set_Draw_color(255, 255, 255);
    lcd.Draw_Round_Rectangle(5, 5, (lcd_size_x/2), (lcd_size_y-6), 7);
    terminal.WriteString("Terminal:", 35, 15, 2, FG_COLOUR, BG_COLOUR, true);
    lcd.Draw_Line(35, 35, (int16_t) (strlen("Terminal:") * 11 + 35), 35);
    // Draw buttons on right
    //lcd.Draw_Round_Rectangle(180, 5, 300, 40, 7);
    //WriteString("Button 1", 195, 15, 2, FG_COLOUR, BG_COLOUR, true);
    //lcd.Draw_Round_Rectangle(180, 50, 300, 85, 7);
    //WriteString("Button 2", 195, 60, 2, FG_COLOUR, BG_COLOUR, true);
    //delay(5000);

}

void loop()
{
    char serial_str[64] = "";
    //char temp_str[BUFFER_CHARS] = "";
    int  num_chars, i;
    int  i2c_temp;

    // Main state machine
    switch (CurrentState)
    {
        case Update_Terminal:
            // Update Terminal
        {
            terminal.DisplayTerminal();
            //terminal.DisplayBuffer(165, 50);
            CurrentState = Wait_For_Serial;
            break;
        }

        case Wait_For_Serial:
        {
            // Reset buffer
            serial_str[0] = '\0';
            num_chars = Serial.available();
            if (num_chars > 0)
            {
                //delay(50);
                for (i=0; i<num_chars; i++)
                {
                    serial_str[i] = (char) Serial.read();
                }
                serial_str[num_chars] = '\0';   // Null terminate
                terminal.WriteTerminalChars(serial_str);
            }
            CurrentState = Wait_For_Touch;

            // If '1' pressed, service I2C
            if (serial_str[0] == '1')
                CurrentState = Read_From_I2C;
            break;
        }

        case Wait_For_Touch:
        {
            /*
            if (is_pressed())
            {
                //Read_Resistive();
                //terminal.WriteTerminalLine("Touch detected!");
                Serial.print("Touch detected!  x:");
                Serial.print(x);
                Serial.print("  y:");
                Serial.println(y);
                //delay(50);
            }
            */
            CurrentState = Update_Terminal;
            break;
        }

        case Read_From_I2C:
        {
            //sendByte(address_bms, 0x2C);
            //sendBytePEC(address_bms, 0x2C);
            //sendByte(address_bms, ABSOLUTE_SOC);
            //i2c_temp = fetchWord(address_bms);
            //sprintf(temp_str, "Received word: 0x%02X", i2c_temp);
            //Serial.println(temp_str);
            //terminal.WriteTerminalLine(temp_str);
            CurrentState = Update_Terminal;
            break;
        }

        default:
        {
            // Should never get here!
            Serial.println("ERROR: Default case hit!");
            CurrentState = Update_Terminal;
        }
    }
}
